"""
Main entrypoint for the cli.

Author: Valtteri Rajalainen
"""

import sys
import typing as types

import templateman
import templateman.cli as cli


def main(args: types.List[str]) -> int:
    if len(args) < 1:
        cli.print_help(None)
        return 0
    
    command = args[0]
    if command not in cli.commands:
        templateman.print_error(f"Unknown command '{command}'. Use 'help' to check all commands...")
        return 1
    
    templateman.running = True
    cli.exec_command(command, args[1:])
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



"""
Utilities for creating template scripts.

Author: Valtteri Rajalainen
"""

import os
import sys
import shutil
import subprocess as subp
import tempfile
import typing as types


running = False
working_dir: str = os.getcwd()
template_info: types.Dict[str, types.Optional[str]] = {
    'name':             'UNKNOWN',
    'output_directory': working_dir,
    'author':           'UNKNOWN',
}


def require_arguments(*args):
    for arg in args:
        if template_info.get(arg) == 'UNKNOWN':
            print_error(f"Missing required argument '{arg}'")
            abort()


def print_error(message: str, add_newline = True):
    prefix = '[ ERROR ] '
    sys.stderr.write(prefix)
    sys.stderr.write(message)
    if add_newline:
        sys.stderr.write('\n')


def abort():
    global running
    if running:
        running = False
        sys.exit(1)


def create_directory(path: str, create_dirs = False):
    try:
        if create_dirs:
            os.makedirs(path)
        else:
            os.mkdir(path)
    
    except (OSError, PermissionError) as err:
        print_error(str(err))
        abort()


def create_file(path: str):
    try:
        open(path, 'x').close()
    except (OSError, PermissionError) as err:
        print_error(str(err))
        abort()


def copy_item(src: str, dst: str):
    try:
        if os.path.isdir(src):
            shutil.copytree(src, dst)
        else:
            shutil.copy(src, dst)

    except (OSError, PermissionError) as err:
        print_error(str(err))
        abort()


def run_command(cmd: types.List[str], path = working_dir) -> types.Tuple[int, str]:
    with tempfile.TemporaryFile(mode='w+') as file:
        proc = subp.run(cmd, cwd=path, stdout=file, stderr=file, text=True)
        file.seek(0)
        return proc.returncode, file.read()




"""
The command line interface implementation.

Author: Valtteri Rajalainen
"""

import sys
import os
import shutil
import runpy
import contextlib
import typing as types
import templateman

VERSION = '1.0.0'
TEMPLATE_DIRECTORY_ENV_VAR = 'PY_TEMPLATES_DIR'

commands: types.Dict[str, types.Callable[[types.List[str],], None]] = dict()


def mkdir_exc_safe(path: str) -> types.Optional[str]:
    error = None
    try:
        os.mkdir(path)
    except Exception as err:
        error = f"{err.__class__.__name__}: {str(err)}"
    return error


def copy_file_exc_safe(src: str, dst: str) -> types.Optional[str]:
    error = None
    try:
        shutil.copy(src, dst)
    except Exception as err:
        error = f"{err.__class__.__name__}: {str(err)}"
    return error


def list_directory_exc_safe(path: str) -> types.Tuple[types.List[str], types.Optional[str]]:
    error = None
    items = list()
    try:
        items = os.listdir(path)
    except Exception as err:
        error = f"{err.__class__.__name__}: {str(err)}"
    return items, error


def open_file_exc_safe(path: str, mode = 'r', *args) -> types.Tuple[object, types.Optional[str]]:
    file = None
    error = None
    try:
        file = open(path, mode, *args)
    except Exception as err:
        error = f"{err.__class__.__name__}: {str(err)}"
    return file, error


def remove_file_exc_safe(path: str) -> types.Optional[str]:
    try:
        os.remove(path)
        return None
    except (OSError, PermissionError) as err:
        return str(err)


def register_command(alias: str):
    def wrapper(func: types.Callable[[types.List[str],], None]):
        commands[alias] = func
        return func
    return wrapper


def exec_command(command: str, args: types.List[str]):
    if command not in commands:
        raise RuntimeError(f"Unknown comman '{command}'")
    commands[command](args)


def parse_args(args: types.List[str], options: dict):
    skip = -1
    for i in range(0, len(args)):
        if skip > i:
            continue
        
        arg = args[i]
        if arg not in options:
            templateman.print_error(f"Unknown option '{arg}'")
            templateman.abort()
            return
        
        args_consumed, handle_options = options[arg]
        start = i + 1
        end = start + args_consumed
        if len(args) < end:
            templateman.print_error(f"Option '{arg}' expected additional arguments")
            templateman.abort()
            return
        
        handle_options(args[start:end])
        skip = end


def resolve_template_directory() -> types.Optional[str]:
    path = None
    with contextlib.suppress(Exception):
        default_directory = os.path.join(os.path.expanduser('~'), '.py-templates')
        path =  os.environ.get(TEMPLATE_DIRECTORY_ENV_VAR, default_directory)
    return path


@register_command('help')
def print_help(args: types.List[str]):
    """Show this help information."""
    def format_command_info(name: str, docstring: str = None) -> str:
        if docstring is None:
            return '  > ' + name + ': ' + 'No description available...\n'
        return '  > ' + name + ': ' + docstring.strip() + '\n'

    commands_help = [format_command_info(name, func.__doc__) for name, func in commands.items()]
    help_text = [
        f'TemplateManager, version: {VERSION}',
        '',
        'A simple utility to manage and run Python scripts for automating',
        'project creation, boilerplate code and much more.'
        '',
        'Usage: ',
        '',
        '   $ python -m templateman [command] [arguments]',
        '',
        '',
        'Implemented commands:',
        '',
        *commands_help,
    ]
    print('\n'.join(help_text))


@register_command('remove')
def remove_installed_template(args: types.List[str]):
    """
    Remove installed template. This will remove the file from the
    template directory permanently.

    USAGE:
    
        $ python -m templateman remove [template-name]

    """
    template_dir = resolve_template_directory()
    if template_dir is None:
        error_message = 'Can\'t resolve users home directory for storing template scripts'
        templateman.print_error(error_message)
        templateman.abort()
        return

    if len(args) == 0:
        templateman.print_error("Command 'remove' expected atleast one argument")
        templateman.abort()
        return

    template_path = os.path.join(template_dir, args[0])
    if not os.path.exists(template_path):
        templateman.print_error(f"No template installed with name '{args[0]}'")
        templateman.abort()
        return

    print(f"Are you sure you want to remove file '{template_path}' premanently?")
    ans = input('Input Y/y to remove this file: ')
    if ans.lower() == 'y':
        error = remove_file_exc_safe(template_path)
        if error:
            error_message = 'Unexpected error when removing file:\n'
            error_message += error
            templateman.print_error(error_message)
            templateman.abort()
            return


@register_command('list')
def list_installed_templates(args: types.List[str]):
    """List all installed templates."""
    template_dir = resolve_template_directory()
    if template_dir is None:
        error_message = 'Can\'t resolve users home directory for storing template scripts'
        templateman.print_error(error_message)
        templateman.abort()
        return

    if not os.path.exists(template_dir):
        error = mkdir_exc_safe(template_dir)
        if error:
            error_message = 'Failed to create directory for storing template scripts:\n'
            error_message += error
            templateman.print_error(error_message)
            templateman.abort()
            return

    installed_templates, error = list_directory_exc_safe(template_dir)
    if error:
        error_message = 'Unexpected error when listing installed templates:\n'
        error_message += error
        templateman.print_error(error_message)
        templateman.abort()
        return

    print('Templates stored in directory:')
    print(template_dir)
    print()
    for template_name in installed_templates:
        print('> ', template_name)
    if installed_templates:
        print()


@register_command('install')
def install_template(args: types.List[str]):
    """
    Install a template script. This creates a copy of the provided file
    into the template directory.

    USAGE:

        $ python -m templateman install [filepath]
    
    """
    template_dir = resolve_template_directory()
    if template_dir is None:
        error_message = 'Can\'t resolve users home directory for storing template scripts'
        templateman.print_error(error_message)
        templateman.abort()
        return
    
    if not os.path.exists(template_dir):
        error = mkdir_exc_safe(template_dir)
        if error:
            error_message = 'Failed to create directory for storing template scripts:\n'
            error_message += error
            templateman.print_error(error_message)
            templateman.abort()
            return
    
    if len(args) < 1:
        templateman.print_error("Command 'install' expected atleast one argument")
        templateman.abort()
        return

    filepath = args[0]
    if not os.path.isabs(filepath):
        filepath = os.path.join(templateman.working_dir, filepath)
    
    if not os.path.exists(filepath):
        templateman.print_error("Given path does not exist")
        templateman.abort()
        return

    _, filename = os.path.split(filepath)
    filename, _ = os.path.splitext(filename)

    error = copy_file_exc_safe(filepath, os.path.join(template_dir, filename))
    if error:
        error_message = 'Install failed:\n'
        error_message += error
        templateman.print_error(error_message)
        templateman.abort()
        return


@register_command('run')
def run_template(args: types.List[str]):
    """
    Execute a template script. If the given filepath has no suffix, an installed
    template is searched. If file has any suffix, the current working directory is
    searched.

    USAGE:

        $ python -m templateman run [template-name] [arguments]

    ARGUMENTS:
        -o / --output-directory: Provide a output directory path for the script.
                                 Default value is the current working directory.

        -n / --name: Provide a name for the script. This should be used to name
                     the generated file, project or other assets.
                     Default value is "UNKNOWN".

        -a / --author: Provide author name for the script. This should be used
                       in config/setup files in place of project's author name.
                       Default value is "UNKNOWN".

    """
    if len(args) < 1:
        templateman.print_error("Command 'run' expected atleast one argument")
        templateman.abort()
        return
    
    filename, suffix = os.path.splitext(args[0])
    filepath = os.path.join(templateman.working_dir, args[0])
    
    if suffix == '':
        template_dir = resolve_template_directory()
        if template_dir is None:
            error_message = 'Can\'t resolve users home directory for storing template scripts'
            templateman.print_error(error_message)
            templateman.abort()
            return
        
        installed_templates, error = list_directory_exc_safe(template_dir)
        if not error and filename in installed_templates:
            filepath = os.path.join(template_dir, filename)

    def set_name(args: types.List[str]):
        templateman.template_info['name'] = args[0]

    def set_author(args: types.List[str]):
        templateman.template_info['author'] = args[0]

    def set_output_directory(args: types.List[str]):
        templateman.template_info['output_directory'] = args[0]

    all_options = {
        '--name': (1, set_name),
        '-n': (1, set_name),

        '--author': (1, set_author),
        '-a': (1, set_author),

        '--output-directory': (1, set_output_directory),
        '-o': (1, set_output_directory),
    }
    parse_args(args[1:], all_options)

    if not os.path.exists(filepath):
        error_message = f"Can't find file '{filepath}'"
        templateman.print_error(error_message)
        templateman.abort()
        return

    file, error = open_file_exc_safe(filepath, 'r')
    if error:
        error_message = f"Can't open file '{filepath}'"
        templateman.print_error(error_message)
        templateman.abort()
        return

    try:
        code = file.read() # type: ignore
        exec(compile(code, filename, 'exec'))
    
    except Exception as err:
        error_message = 'There were errors during the execution of the script:'
        error_message += f'\n{err.__class__.__name__}: {str(err)}'
        templateman.print_error(error_message)
        templateman.abort()
    
    finally:
        file.close() # type: ignore



from setuptools import find_packages, setup


LONG_DESCRIPTION = """# templateman
A simple project template manager written in Python.

Create Python scripts to automate boilerplate code creation,
project templates, or other tasks. Templateman provides a simple
CLI for managing these scripts.

"""


setup(
    name = 'templateman',
    version = '1.0.0',
    description = 'A simple project template manager written in Python.',
    long_description = LONG_DESCRIPTION,
    long_description_content_type = "text/markdown",
    url = "https://github.com/varajala/templateman",
    
    author = 'Valtteri Rajalainen',
    author_email = 'rajalainen.valtteri@gmail.com',

    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        
        "Topic :: Software Development",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
    ],
    
    python_requires = '>=3.7',
    packages = find_packages(),
)
